package algorithms;

import java.util.Vector;

import main.Main;
import structures.Node;
import structures.Solution;
import structures.Table;
import structures.Zone;

public class IDAStar extends Algorithm {

	private Integer nodes;
	private Integer cuts;
	private Integer bound;
	private Long time;
	
	public Solution solve() {
		time = new Long(System.currentTimeMillis());
		nodes = new Integer(1);
		cuts = new Integer(0);
		bound = new Integer(Main.heuristic.calculate(Main.table));
		Solution solution = new Solution(-1, (int) (System.currentTimeMillis() - time), nodes, cuts);
		while(solution.getLevel() == -1 && !logicStop) solution = cutOffSolve();
		return solution;
	}
	
	private Solution cutOffSolve() {
		Vector<Node> frontier = new Vector<Node>();
		Node currentNode;
		Node evaluationNode;
		Table evaluationTable;
		Zone legalMoves;
		Vector<Integer> childrenBounds = new Vector<Integer>();
		frontier.add(0, new Node(Main.table, null, 0, Main.heuristic.calculate(Main.table)));
		while(!frontier.elementAt(0).getStatus() && frontier.elementAt(0).getValue() <= bound && !logicStop) {
			synchronized(graphicStop) {logicStop = new Boolean(graphicStop);}
			currentNode = frontier.elementAt(0);
			frontier.addElement(frontier.remove(0));
			currentNode.setStatus(true);
			if(goalTest(currentNode.getTable())) 
				return new Solution(currentNode.getLevel(), (int) (System.currentTimeMillis() - time), nodes, cuts);
			legalMoves = Algorithm.legalMoves(currentNode.getTable());
			for(int i = 0; i < legalMoves.getSize(); i++) {
				evaluationTable = (currentNode.getTable()).copy();
				evaluationTable.switchTokens(legalMoves.getZeroLine(), legalMoves.getZeroColumn(), legalMoves.getLine(i), legalMoves.getColumn(i));
				evaluationNode = new Node(evaluationTable, currentNode, currentNode.getLevel() + 1, currentNode.getLevel() + 1 + Main.heuristic.calculate(evaluationTable));
				if(Main.policy.policyCheck(currentNode, evaluationNode) && evaluationNode.getValue() <= bound) {
					nodes++;
					currentNode.setSon(evaluationNode);
					int j = 0;
					while(j < frontier.size() && !frontier.elementAt(j).getStatus() && frontier.elementAt(j).getValue() < evaluationNode.getValue()) j++;
					frontier.add(j, evaluationNode);
				}
				else {
					if(evaluationNode.getValue() > bound) childrenBounds.addElement(new Integer(evaluationNode.getValue()));
					else cuts++;
					}
			}
		}
		bound = childrenBounds.elementAt(0);
		for(int i = 1; i < childrenBounds.size(); i++) {
			if(childrenBounds.elementAt(i) < bound) bound = childrenBounds.elementAt(i);
			}
		return new Solution(-1, (int) (System.currentTimeMillis() - time), nodes, cuts);
	}
}
