package algorithms;

import main.Main;
import structures.Node;
import structures.Solution;
import structures.Table;
import structures.Zone;

//Questo algoritmo assieme all'euristica Manhattan dà risultati di valore maggiorato

public class RecursiveBF extends Algorithm {

	private Integer nodes;
	private Integer cuts;
	private Long time;
	
	public Solution solve() {
		time = new Long(System.currentTimeMillis());
		nodes = new Integer(1);
		cuts = new Integer(0);
		int alternative = 1;
		for(int i = 2; i <= Main.table.getSize() * Main.table.getSize(); i++) alternative = alternative * i;
		return RBFSolve(new Node(Main.table, null, 0, Main.heuristic.calculate(Main.table)), alternative);
	}
	
	private Solution RBFSolve(Node father, int alternative) {
		Node evaluationNode;
		Solution innerSolution;
		Node best;
		Table evaluationTable;
		Zone legalMoves;
		if(goalTest(father.getTable())) 
			return new Solution(father.getLevel(), (int) (System.currentTimeMillis() - time), nodes, cuts);
		legalMoves = Algorithm.legalMoves(father.getTable());
		for(int i = 0; i < legalMoves.getSize(); i++) {
			evaluationTable = (father.getTable()).copy();
			evaluationTable.switchTokens(legalMoves.getZeroLine(), legalMoves.getZeroColumn(), legalMoves.getLine(i), legalMoves.getColumn(i));
			evaluationNode = new Node(evaluationTable, father, father.getLevel() + 1, Math.max(father.getLevel() + 1 + Main.heuristic.calculate(evaluationTable), father.getValue()));
			if(Main.policy.policyCheck(father, evaluationNode) && evaluationNode.getValue() <= alternative) {
				nodes++;
				father.setSon(evaluationNode);
			}
			else if(evaluationNode.getValue() > alternative) cuts++;
			}
		if(father.getSonSize() == 0) return new Solution(-1, (int) (System.currentTimeMillis() - time), nodes, cuts);
		while(!logicStop) {
			synchronized(graphicStop) {logicStop = new Boolean(graphicStop);}
			best = firstBestValue(father);
			if(best.getValue() > alternative) return new Solution(-1, (int) ((System.currentTimeMillis() - time)/1000), nodes, best.getValue());
			int limit = secondBestValue(father).getValue();
			innerSolution = RBFSolve(best, Math.min(limit, alternative));
			if(innerSolution.getLevel() == -1) best.setValue(innerSolution.getCut());
			else return innerSolution;
			}
		return new Solution(-1, (int) (System.currentTimeMillis() - time), nodes, cuts);
		}
	
	private Node firstBestValue(Node father) {
		Node best = father.getSon(0);
		for(int i = 1; i < father.getSonSize(); i++) {
			if(father.getSon(i).getValue() < best.getValue()) best = father.getSon(i);
		}
		return best;
	}
	
	private Node secondBestValue(Node father) {
		Node first = firstBestValue(father);
		if(father.getSonSize() == 1) return first;
		Node second;
		if(father.getSon(0) != first) second = father.getSon(0);
		else second = father.getSon(1);
		for(int i = 0; i < father.getSonSize(); i++) {
			if(father.getSon(i).getValue() < second.getValue() && first != father.getSon(i)) second = father.getSon(i);
		}
		return second;
	}
}
